from matplotlib import pyplot as plt
import matplotlib.cm as cm
import matplotlib.gridspec as gridspec
import pandas as pd
import numpy as np
import math as math
from adjustText import adjust_text

from . import tree
from . import per
from . import des
from . import notes
from . import prob
from . import samp  # sampling
from . import hyp  # hypothesis
# from . import t  # hypothesis using t instead of z

from ._version import __version__


def geomean(rate):
    """
    Result: Will return the geometric mean

    Usage:
    return_rate = np.array([0.2, 0.1, -0.05])
    avg_rate = geomean(return_rate)
    print("geo_mean =", avg_rate)
    """
    rate_1 = rate + 1
    geo_m = math.exp(np.log(rate_1).mean()) - 1
    return geo_m


def ori_autolabel(rects, truncate='{}'):
    """Attach a text label above each bar in *rects*, displaying its height."""
    for rect in rects:
        height = rect.get_height()
        plt.annotate(truncate.format(height),
                     xy=(rect.get_x() + rect.get_width() / 2, height),
                     xytext=(0, 3),  # 3 points vertical offset
                     textcoords="offset points",
                     ha='center', va='bottom')


def autolabel(rects, **kwargs):
    """
    Attach a text label above each bar in *rects*, displaying its height.

    list of kwargs:
        + bool original
        + string truncate: in the '{}' format. Can be filled in with '{:.5f}'
        + dict arrowprops:
            usage: pass in `arrowprops=dict(arrowstyle='->', color='r')`
    """
    if "original" in kwargs and kwargs["original"] == True:
        if "truncate" in kwargs:
            ori_autolabel(rects, kwargs["truncate"])
            return
        else:
            ori_autolabel(rects)
        return

    texts = []
    for rect in rects:
        height = rect.get_height()
        if "truncate" in kwargs:
            texts.append(plt.text(s=kwargs["truncate"].format(height),
                                  x=(rect.get_x() + rect.get_width() / 2),
                                  y=(height)))
        else:
            texts.append(plt.text(s='{}'.format(height),
                                  x=(rect.get_x() + rect.get_width() / 2),
                                  y=(height)))

    if "arrowprops" in kwargs:
        adjust_text(texts, arrowprops=kwargs['arrowprops'])


def color_palette(n, cmap="jet"):
    '''
    Generate a series of color using matplotlib color map templates.
    The default color map (cmap) is set to "jet".
    '''

    get_color = getattr(cm, cmap)
    return get_color(np.linspace(0, 1, n))


def add_margin(ax, x=0.05, y=0.05):
    '''
     Setting margins in matplotlib/seaborn with subplots.

     This will, by default, add 5% to the x and y margins. You
     can customize this using the x and y arguments when you call it.
     ref: https://stackoverflow.com/a/34205235/10871988

     e.g.
     ax = sns.regplot(x='Temperature', y= 'Tickets', data = df, ci = None, scatter_kws={'color':'dodgerblue'}, line_kws={'linestyle':'dashed', 'color':'#ffaa77'})
     add_margin(ax, x=0.01, y=0.00) ### Call this after regplot 
    '''

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    xmargin = (xlim[1]-xlim[0])*x
    ymargin = (ylim[1]-ylim[0])*y

    ax.set_xlim(xlim[0]-xmargin, xlim[1]+xmargin)
    ax.set_ylim(ylim[0]-ymargin, ylim[1]+ymargin)


class Pos():
    # positioning - src: http://emptypipes.org/2013/11/09/matplotlib-multicategory-barchart/
    def __init__(self, groupedN, seriesN, space=0.3):
        self.space = space
        self.groupedN = groupedN
        self.seriesN = seriesN
        self.width = (1 - self.space) / (self.groupedN)
        self.__calcPos()

    def __calcPos(self):
        self.pos = list()
        for i in range(self.seriesN):
            self.pos.append([j - (1 - self.space) / 2. + i *
                             self.width for j in range(1, self.groupedN+1)])

    def returnPos(self):
        return self.pos

    def restructure(self, groupedN, seriesN):
        self.groupedN = groupedN
        self.seriesN = seriesN
        self.width = (1 - self.space) / (self.groupedN)
        self.__calcPos()



from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
import math
import scipy.stats as stats

from .. import samp


def samp_size(s_p, width, alpha):
    z_cv = stats.norm.ppf(1 - alpha / 2)
    return (z_cv * math.sqrt(s_p * (1 - s_p)) / width) ** 2


def con_level(s_p, n, alpha, show=True, Wilson=False, N=False):
    """
    Caution: np̂ > 5 and n(1 - p̂) > 5
    """
    con_coef = 1 - alpha
    z_cv = stats.norm.ppf(1 - alpha / 2)
    if not Wilson and not samp.check5(n, s_p):
        print('Not satisfying np̂ > 5 and n(1 - p̂) > 5...')
    if Wilson:
        # make sure that you have arrange s_p to (x + 2) / (n + 4)
        lat = z_cv * math.sqrt(s_p * (1 - s_p)/(n + 4))
    else:
        lat = z_cv * math.sqrt(s_p * (1 - s_p)/n)

    lcl = s_p - lat
    ucl = s_p + lat

    if N:
        if n / N > 0.5:
            fpcf = math.sqrt((N - n)/(N - 1))
            lcl = s_p - lat * fpcf
            ucl = s_p + lat * fpcf
        if lcl < 0:
            lcl = 0
        if ucl < 0:
            ucl = 0
        result = f"""{con_coef * 100:.1f}% Confidence Interval: N [{lcl:.4f}, {ucl:.4f}] = [{N * lcl:.4f}, {N * ucl:.4f}]
p̂: {s_p:.4f}
Sample Size: {n}
z_cv (Critical value): {z_cv:.4f}
    """
    else:
        # if lcl < 0:
        #     lcl = 0
        # if ucl < 0:
        #     ucl = 0
        result = f"""{con_coef * 100:.1f}% Confidence Interval: [{lcl:.4f}, {ucl:.4f}]
p̂: {s_p:.4f}
Sample Size: {n}
z_cv (Critical value): {z_cv:.4f}
    """
    if show:
        print(result)
    return {"lcl": lcl, "ucl": ucl, "z_cv": z_cv}


def rejection_region_method(s_p, h0_p, nsize, alpha, option='left', precision=4, show=True, ignore=False):
    opt = option.lower()[0]
    if not samp.check5(nsize, h0_p):
        print('Not satisfying np_0 > 5 and n(1 - p_0) > 5...')
    if opt == 't':
        option = 'Two-Tail Test'
        z_cv = stats.norm.ppf(1 - alpha/2)
        p_u = h0_p + z_cv * math.sqrt(h0_p * (1 - h0_p)/nsize)
        p_l = h0_p - z_cv * math.sqrt(h0_p * (1 - h0_p)/nsize)
        flag = s_p < p_l or s_p > p_u
        if not ignore:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z (Critical value) = {z_cv:.{precision}f}

Using {option}:
p̂ =  {s_p:.{precision}f}
p_l (Lower bound for the critical value) = {p_l:.{precision}f}
p_u (Upper bound for the critical value) = {p_u:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z (Critical value) = {z_cv:.{precision}f}

Using {option}:
p_l (Lower bound for the critical value) = {p_l:.{precision}f}
p_u (Upper bound for the critical value) = {p_u:.{precision}f}
            '''

    else:
        z_cv = stats.norm.ppf(1 - alpha)
        if opt == 'l':
            # left tail
            option = 'One-Tail Test (left tail)'
            p_c = h0_p - z_cv * math.sqrt(h0_p * (1 - h0_p)/nsize)
            flag = s_p < p_c
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            p_c = h0_p + z_cv * math.sqrt(h0_p * (1 - h0_p)/nsize)
            flag = s_p > p_c
        if not ignore:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z (Critical value) = {z_cv:.{precision}f}

Using {option}:
p̂ =  {s_p:.{precision}f}
p_c (Critical value) = {p_c:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z (Critical value) = {z_cv:.{precision}f}

Using {option}:
p_c (Critical value) = {p_c:.{precision}f}
            '''

    if show:
        print(result)

    if opt == 't':
        return p_l, p_u
    else:
        return p_c


def testing_statistic_method(s_p, h0_p, nsize, alpha, option='left', precision=4, ignore=False):
    opt = option.lower()[0]
    z_stats = (s_p - h0_p)/math.sqrt(h0_p * (1 - h0_p)/nsize)
    if not samp.check5(nsize, h0_p):
        print('Not satisfying np_0 > 5 and n(1 - p_0) > 5...')

    if opt == 't':
        z_cv = stats.norm.ppf(1 - alpha / 2)
        option = 'Two-Tail Test'
        flag = z_stats < -z_cv or z_stats > z_cv

        if not ignore:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z_cv (Critical value) = {z_cv:.{precision}f}

Using {option}:
z_stats (Observed value) =  {z_stats:.{precision}f}
-z_cv (Lower bound for the critical value) = {-z_cv:.{precision}f}
z_cv (Upper bound for the critical value) = {z_cv:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z_cv (Critical value) = {z_cv:.{precision}f}

Using {option}:
-z_cv (Lower bound for the critical value) = {-z_cv:.{precision}f}
z_cv (Upper bound for the critical value) = {z_cv:.{precision}f}
            '''

    else:
        z_cv = stats.norm.ppf(1 - alpha)
        if opt == 'l':
            # left tail
            option = 'One-Tail Test (left tail)'
            z_cv = -z_cv
            flag = z_stats < z_cv
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            flag = z_stats > z_cv

        if not ignore:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z_cv (Critical value) = {z_cv:.{precision}f}

Using {option}:
z_stats (Observed value) =  {z_stats:.{precision}f}
z_cv (Critical value) = {z_cv:.{precision}f}
Reject H_0 → {flag}
            '''

        else:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}
z_cv (Critical value) = {z_cv:.{precision}f}

Using {option}:
z_cv (Critical value) = {z_cv:.{precision}f}
            '''

    print(result)

    return z_stats, z_cv


def inter_p_value(p_value):
    # interpretation
    if p_value >= 0 and p_value < 0.01:
        inter_p = 'Overwhelming Evidence'
    elif p_value >= 0.01 and p_value < 0.05:
        inter_p = 'Strong Evidence'
    elif p_value >= 0.05 and p_value < 0.1:
        inter_p = 'Weak Evidence'
    elif p_value >= .1:
        inter_p = 'No Evidence'
    return inter_p


def p_value_method(s_p, h0_p, nsize, alpha, option='left', precision=4):
    opt = option.lower()[0]
    z_stats = (s_p - h0_p)/math.sqrt(h0_p * (1 - h0_p)/nsize)
    if not samp.check5(nsize, h0_p):
        print('Not satisfying np_0 > 5 and n(1 - p_0) > 5...')
    if opt == 't':
        # two-tail test
        option = 'Two-Tail Test'
        if s_p > h0_p:
            p_value = stats.norm.sf(z_stats) * 2
        else:
            p_value = stats.norm.cdf(z_stats) * 2

        z_cv = stats.norm.ppf(1 - alpha/2)
        flag = p_value < alpha
        sub_result = f'''Using {option}:
Difference = {s_p - h0_p}
z_cv (Critical value) = {-z_cv:.{precision}f}, {z_cv:.{precision}f}
z_stats (Observed value) = {z_stats:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 → {flag}
        '''
    else:
        if opt == 'l':
            option = 'One-Tail Test (left tail)'
            p_value = stats.norm.cdf(z_stats)
            z_cv = -stats.norm.ppf(1 - alpha)
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            p_value = stats.norm.sf(z_stats)
            z_cv = stats.norm.ppf(1 - alpha)
        flag = p_value < alpha
        sub_result = f'''Using {option}:
Difference = {s_p - h0_p}
z_cv (Critical value) = {z_cv:.{precision}f}
z_stats (Observed value) = {z_stats:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 → {flag}
        '''

    result = f"""======= p-value Method =======
p̂ = {s_p:.{precision}f}
Number of Observation = {nsize:.{precision}f}
Hypothesized Proportion (H0 Mean) = {h0_p:.{precision}f}
Significant Level (alpha) = {alpha:.{precision}f}

""" + sub_result

    print(result)

    return z_cv, z_stats, p_value



from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
import math
import scipy.stats as stats


def con_level(alpha, n, s, sigma, show=True, ignore=True):
    df_v = n - 1
    con_coef = 1 - alpha

    chi2_cv_u = stats.chi2.ppf(1 - alpha / 2, df=df_v)
    chi2_cv_l = stats.chi2.ppf(alpha / 2, df=df_v)

    lcl = (df_v) * (s ** 2) / chi2_cv_u
    ucl = (df_v) * (s ** 2) / chi2_cv_l

    result = f"""{con_coef * 100:.1f}% Confidence Interval: [{lcl:.4f}, {ucl:.4f}]

S^2 (Sample Variance): {s ** 2:.4f}
Sample Size: {n}
"""
    if ignore == False:
        flag = sigma ** 2 >= lcl and sigma ** 2 <= ucl
        if flag:
            cover = "covers the real/population variance."
        else:
            cover = "does not cover the real/population variance."

        result += f'''sigma^2 (Population Variance) = {sigma ** 2:.4f}
The interval {cover}
'''
    if show:
        print(result)
    return {"lcl": lcl, "ucl": ucl}


def rejection_region_method(alpha, n, s, sigma, option='left', precision=4, show=True, ignore=False):
    opt = option.lower()[0]
    df_v = n - 1
    var = s ** 2
    if opt == 't':
        option = 'Two-Tail Test'
        chi2_cv_u = stats.chi2.ppf(1 - alpha / 2, df=df_v)
        chi2_cv_l = stats.chi2.ppf(alpha / 2, df=df_v)
        s2_u = chi2_cv_u * (sigma ** 2) / df_v
        s2_l = chi2_cv_l * (sigma ** 2) / df_v
        flag = var < s2_l or var > s2_u
        if not ignore:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2_l (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
chi2_u (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}

Using {option}:
S^2 (var) =  {var:.{precision}f}
S_L^2 (Lower bound for the critical value) = {s2_l:.{precision}f}
S_U^2 (Upper bound for the critical value) = {s2_u:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2_l (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
chi2_u (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}

Using {option}:
S_L^2 (Lower bound for the critical value) = {s2_l:.{precision}f}
S_U^2 (Upper bound for the critical value) = {s2_u:.{precision}f}
            '''

    else:
        if opt == 'l':
            # left tail
            option = 'One-Tail Test (left tail)'
            chi2_value = df_v * (s ** 2) / (sigma ** 2)
            chi2_cv = stats.chi2.ppf(alpha, df_v)
            s2_c = chi2_cv * (sigma ** 2) / df_v
            flag = var < s2_c
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            chi2_value = df_v * (s ** 2) / (sigma ** 2)
            chi2_cv = stats.chi2.ppf(1 - alpha, df_v)
            s2_c = chi2_cv * (sigma ** 2) / df_v
            flag = var > s2_c
        if not ignore:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2 (Critical value) = {chi2_cv:.{precision}f}

Using {option}:
S^2 (var) =  {var:.{precision}f}
S_C^2 (Critical value) = {s2_c:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2 (Critical value) = {chi2_cv:.{precision}f}

Using {option}:
S_C^2 (Critical value) = {s2_c:.{precision}f}
            '''

    if show:
        print(result)

    if opt == 't':
        return s2_l, s2_u
    else:
        return s2_c


def testing_statistic_method(alpha, n, s, sigma, option='left', precision=4, ignore=False):
    opt = option.lower()[0]
    df_v = n - 1
    var = s ** 2
    chi2_value = df_v * (s ** 2) / (sigma ** 2)

    if opt == 't':
        option = 'Two-Tail Test'
        chi2_cv_u = stats.chi2.ppf(1 - alpha / 2, df=df_v)
        chi2_cv_l = stats.chi2.ppf(alpha / 2, df=df_v)
        flag = chi2_value < chi2_cv_l or chi2_value > chi2_cv_u

        if not ignore:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2 (Observed value) =  {chi2_value:.{precision}f}
Chi2_L (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
Chi2_U (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2_L (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
Chi2_U (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}
            '''

    else:
        if opt == 'l':
            # left tail
            option = 'One-Tail Test (left tail)'
            chi2_cv = stats.chi2.ppf(alpha, df_v)
            flag = chi2_value < chi2_cv
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            chi2_cv = stats.chi2.ppf(1 - alpha, df_v)
            flag = chi2_value > chi2_cv

        if not ignore:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2 (Observed value) =  {chi2_value:.{precision}f}
Chi2_C (Critical value) = {chi2_cv:.{precision}f}
Reject H_0 → {flag}
            '''

        else:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2_C (Critical value) = {chi2_cv:.{precision}f}
            '''

    print(result)
    if opt == 't':
        return chi2_value, chi2_cv_l, chi2_cv_u
    else:
        return chi2_value, chi2_cv


def inter_p_value(p_value):
    # interpretation
    if p_value >= 0 and p_value < 0.01:
        inter_p = 'Overwhelming Evidence'
    elif p_value >= 0.01 and p_value < 0.05:
        inter_p = 'Strong Evidence'
    elif p_value >= 0.05 and p_value < 0.1:
        inter_p = 'Weak Evidence'
    elif p_value >= .1:
        inter_p = 'No Evidence'
    return inter_p


def p_value_method(alpha, n, s, sigma, option='left', precision=4):
    opt = option.lower()[0]
    df_v = n - 1
    var = s ** 2
    chi2_stat = df_v * (s ** 2) / (sigma ** 2)

    if opt == 't':
        # two-tail test
        option = 'Two-Tail Test'
        alphas = np.arange(0, 1, 0.001)
        chi2s = stats.chi2.ppf(alphas, df_v)
        med = np.median(chi2s)

        if chi2_stat > med:
            p_value = 2 * stats.chi2.sf(chi2_stat, df_v)
        else:
            p_value = 2 * stats.chi2.cdf(chi2_stat, df_v)

        flag = p_value < alpha
        sub_result = f'''Using {option}:
Chi2 Median under d.f. {df_v}: {med:.{precision}f}
Chi2 Stat (Observed value) = {chi2_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 → {flag}
        '''
    else:
        if opt == 'l':
            option = 'One-Tail Test (left tail)'
            p_value = stats.chi2.cdf(chi2_stat, df_v)
            chi2_cv = stats.chi2.ppf(alpha, df_v)
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            p_value = stats.chi2.sf(chi2_stat, df_v)
            chi2_cv = stats.chi2.ppf(1 - alpha, df_v)
        flag = p_value < alpha
        sub_result = f'''Using {option}:
Chi2_C (Critical value) = {chi2_cv:.{precision}f}
Chi2 Stat (Observed value) = {chi2_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 → {flag}
        '''

    result = f"""======= p-value Method =======
S^2 (var) = {var:.{precision}f}
Number of Observation = {n:.{precision}f}
Hypothesized Std. (H0 Std.) = {sigma:.{precision}f}
Sample Standard Deviation = {s:.{precision}f}
Significant Level (alpha) = {alpha:.{precision}f}

""" + sub_result

    print(result)

    return chi2_stat, p_value
